use lazy_static::lazy_static;
use x86_64::structures::gdt::{Descriptor, GlobalDescriptorTable, SegmentSelector};
use x86_64::structures::tss::TaskStateSegment;
use x86_64::VirtAddr;

// struct Selectors ::: selectors returned by GDT
// code_selector: SegmentSelector :field: indicates where the CS register
//   needs to be put in case of a double fault
// tss_selector: SegmentSelector :field: indicates where the TSS ought to start
//   in case of a double fault
struct Selectors {
    code_selector: SegmentSelector,
    tss_selector: SegmentSelector,
}

// const DOUBLE_FAULT_IST_INDEX: u16 ::: virtual location of double fault stack
pub const DOUBLE_FAULT_IST_INDEX: u16 = 0;

lazy_static! {
    // GDT: (GlobalDescriptorTable, Selectors) ::: antiquated, used to specify TSS
    // SAFETY ensure the correct selectors are returned
    static ref GDT: (GlobalDescriptorTable, Selectors) = {
        let mut gdt = GlobalDescriptorTable::new();
        let code_selector = gdt.add_entry(Descriptor::kernel_code_segment());
        let tss_selector = gdt.add_entry(Descriptor::tss_segment(&TSS));
        (
            gdt,
            Selectors {
                code_selector,
                tss_selector,
            },
        )
    };

    // TSS: TaskStateSegment ::: holds stack tables
    static ref TSS: TaskStateSegment = {
        let mut tss = TaskStateSegment::new();
        tss.interrupt_stack_table[DOUBLE_FAULT_IST_INDEX as usize] = {
            const STACK_SIZE: usize = 4096 * 5;
            static mut STACK: [u8; STACK_SIZE] = [0; STACK_SIZE]; // TODO make real stack
            let stack_start = VirtAddr::from_ptr(&raw const STACK);
            let stack_end = stack_start + STACK_SIZE;
            stack_end
        };
        tss
    };
}

// fn init ::: initializes and loads GDT and TSS and places CS accordingly
pub fn init() {
    use x86_64::instructions::segmentation::{Segment, CS};
    use x86_64::instructions::tables::load_tss;
    GDT.0.load();
    // SAFETY ensure GDT returns the right values
    unsafe {
        CS::set_reg(GDT.1.code_selector);
        load_tss(GDT.1.tss_selector);
    }
}
