#![no_std]
#![cfg_attr(test, no_main)]
#![feature(abi_x86_interrupt)]
#![feature(custom_test_frameworks)]
#![test_runner(test_runner)]
#![reexport_test_harness_main = "test_main"]

extern crate alloc;

pub mod allocator;
pub mod gdt;
pub mod interrupts;
pub mod memory;
pub mod serial;
pub mod vga_buffer;

#[cfg(test)]
use bootloader::{entry_point, BootInfo};
use core::panic::PanicInfo;

#[cfg(test)]
entry_point!(test_kernel_main);

// trait Testable ::: defines run() for all test functions
pub trait Testable {
    fn run(&self) -> ();
}

impl<T> Testable for T
where
    T: Fn(),
{
    fn run(&self) {
        serial_print!("{}...\t", core::any::type_name::<T>());
        self();
        serial_println!("[ok]");
    }
}

// enum QemuExitCode ::: used when exiting with Qemu's isa-debug-exit device
// Success :field:
// Failed :field:
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(u32)]
pub enum QemuExitCode {
    Success = 0x10,
    Failed = 0x11,
}

// fn test_kernel_main :::
// _boot_info: &'static BootInfo :param:
#[cfg(test)]
fn test_kernel_main(_boot_info: &'static BootInfo) -> ! {
    init();
    test_main();
    hlt_loop();
}

// fn exit_qemu ::: uses Qemu's isa-debug-exit device to quit
// exit_code: QemuExitCode :param:
pub fn exit_qemu(exit_code: QemuExitCode) {
    use x86_64::instructions::port::Port;
    unsafe {
        let mut port = Port::new(0xf4);
        port.write(exit_code as u32);
    }
}

// fn hlt_loop ::: idles the cpu
pub fn hlt_loop() -> ! {
    loop {
        x86_64::instructions::hlt();
    }
}

// fn init ::: initializes everything; ought be the first thing that's called
pub fn init() {
    gdt::init();
    interrupts::init_idt();
    unsafe { interrupts::PICS.lock().initialize() };
    x86_64::instructions::interrupts::enable();
}

// fn test_panic_handler ::: prints error info, exits qemu
// info: &PanicInfo :param:
pub fn test_panic_handler(info: &PanicInfo) -> ! {
    serial_println!("[failed]\n");
    serial_println!("Error: {}\n", info);
    exit_qemu(QemuExitCode::Failed);
    hlt_loop(); // unreachable but Rust doesn't know that
}

// fn test_runner ::: cycles through all Testable
// tests: &[&dyn Testable] :param: every Testable function in scope
pub fn test_runner(tests: &[&dyn Testable]) {
    serial_println!("Running {} tests", tests.len());
    for test in tests {
        test.run();
    }
    exit_qemu(QemuExitCode::Success);
}

// fn panic ::: calls test_panic_handler, which is basically a public wrapper
// info: &PanicInfo :param:
#[cfg(test)]
#[panic_handler]
fn panic(info: &PanicInfo) -> ! {
    test_panic_handler(info)
}
