use core::fmt;
use lazy_static::lazy_static;
use spin::Mutex;
use volatile::Volatile;

// macro print ::: prints to vga_buffer
#[macro_export]
macro_rules! print {
  ($($arg:tt)*) => ($crate::vga_buffer::_print(format_args!($($arg)*)));
}

// macro println ::: prints to vga_buffer with \n appended
#[macro_export]
macro_rules! println {
  () => ($crate::print!("\n"));
  ($($arg:tt)*) => ($crate::print!("{}\n", format_args!($($arg)*)));
}

// enum Color ::: defines color bytes for VGA buffer as u8
#[allow(dead_code)]
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(u8)]
pub enum Color {
    Black = 0,
    Blue = 1,
    Green = 2,
    Cyan = 3,
    Red = 4,
    Magenta = 5,
    Brown = 6,
    LightGray = 7,
    DarkGray = 8,
    LightBlue = 9,
    LightGreen = 10,
    LightCyan = 11,
    LightRed = 12,
    Pink = 13,
    Yellow = 14,
    White = 15,
}

// const BUFFER_HEIGHT :::
// const BUFFER_WIDTH :::
const BUFFER_HEIGHT: usize = 25;
const BUFFER_WIDTH: usize = 80;

// struct Buffer :::
// chars: [[Volatile<ScreenChar>; BUFFER_WIDTH]; BUFFER_HEIGHT] :field: char array of text buffer.
//   Volatile ensures it'll never be optimized away because this buffer is a side effect
#[repr(transparent)]
struct Buffer {
    chars: [[Volatile<ScreenChar>; BUFFER_WIDTH]; BUFFER_HEIGHT],
}

// struct ColorCode ::: formatted for VGA buffer; first 4 foreground, last 4 background
// (u8) :field:
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(transparent)]
struct ColorCode(u8);

impl ColorCode {
    // fn new ::: creates color code to VGA specs
    // foreground: Color :param:
    // background: Color :param:
    // : ColorCode :ret:
    fn new(foreground: Color, background: Color) -> ColorCode {
        ColorCode((background as u8) << 4 | (foreground as u8))
    }
}

// struct ScreenChar ::: specifies character and color of 1 character for VGA buffer
// ascii_character: u8 :field:
// color_code: ColorCode :field:
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(C)]
struct ScreenChar {
    ascii_character: u8,
    color_code: ColorCode,
}

// struct Writer :::
// column_position: usize :field:
// color_code: ColorCode :field:
// buffer: &'static mut Buffer :field: buffer to write to
pub struct Writer {
    column_position: usize,
    color_code: ColorCode,
    buffer: &'static mut Buffer,
}

impl Writer {
    // fn write_byte ::: writes byte to screen
    // &mut self :param: self whose buffer will be written to
    // byte: u8 :param: byte to write to screen. Must be writable
    pub fn write_byte(&mut self, byte: u8) {
        match byte {
            b'\n' => self.new_line(),
            byte => {
                // wraps automatically
                if self.column_position >= BUFFER_WIDTH {
                    self.new_line();
                }

                // bottom row of screen will be written to
                let row = BUFFER_HEIGHT - 1;
                let col = self.column_position;

                let color_code = self.color_code;
                self.buffer.chars[row][col].write(ScreenChar {
                    ascii_character: byte,
                    color_code,
                });

                // moves on to next byte's position preemptively
                self.column_position += 1;
            }
        }
    }

    // fn write_string :::
    // &mut self :param: self whose buffer will be written to
    // s: &str :param: string to write
    pub fn write_string(&mut self, s: &str) {
        for byte in s.bytes() {
            match byte {
                0x20..=0x7e | b'\n' => self.write_byte(byte),
                // placeholder for unknown bytes
                _ => self.write_byte(0xfe),
            }
        }
    }

    // fn new_line ::: shifts buffer contents up, replaces cursor on the left
    // &mut self :param: self whose buffer will be written to
    fn new_line(&mut self) {
        for row in 1..BUFFER_HEIGHT {
            for col in 0..BUFFER_WIDTH {
                let character = self.buffer.chars[row][col].read();
                self.buffer.chars[row - 1][col].write(character);
            }
        }
        self.clear_row(BUFFER_HEIGHT - 1);
        self.column_position = 0;
    }

    // fn clear_row ::: replaces row with spaces
    // &mut self :param: self whose buffer will be written to
    // row: usize :param: row of buffer to clear
    fn clear_row(&mut self, row: usize) {
        let blank = ScreenChar {
            ascii_character: b' ',
            color_code: self.color_code,
        };
        for col in 0..BUFFER_WIDTH {
            self.buffer.chars[row][col].write(blank);
        }
    }
}

impl fmt::Write for Writer {
    // fn write_str :::
    // &mut self :param: self whose buffer will be written to
    // s: &str :param: string to be written
    // : fmt::Result :ret:
    fn write_str(&mut self, s: &str) -> fmt::Result {
        self.write_string(s);
        Ok(())
    }
}

lazy_static! {
  // WRITER: Mutex<Writer> :::
  pub static ref WRITER: Mutex<Writer> = Mutex::new(Writer {
    column_position: 0,
    color_code: ColorCode::new(Color::Magenta, Color::Black),
    buffer: unsafe { &mut *(0xb8000 as *mut Buffer) },
  });
}

// fn _print ::: used for macro magic
// args: fmt::Arguments :param:
#[doc(hidden)]
pub fn _print(args: fmt::Arguments) {
    use core::fmt::Write;
    use x86_64::instructions::interrupts;
    interrupts::without_interrupts(|| {
        WRITER.lock().write_fmt(args).unwrap();
    });
}

// fn test_println_many :::
#[test_case]
fn test_println_many() {
    for _ in 0..1023 {
        println!("test_println_many output");
    }
}

// fn test_println_output ::: checks that what is getting written is getting printed
#[test_case]
fn test_println_output() {
    use core::fmt::Write;
    use x86_64::instructions::interrupts;

    let s = "Some test string that fits in a single line";
    interrupts::without_interrupts(|| {
        // prevent race condition where an exception prints while it io
        let mut writer = WRITER.lock();
        writeln!(writer, "\n{}", s).expect("writeln failed");
        for (i, c) in s.chars().enumerate() {
            let screen_char = writer.buffer.chars[BUFFER_HEIGHT - 2][i].read();
            assert_eq!(char::from(screen_char.ascii_character), c);
        }
    });
}

// fn test_println_simple :::
#[test_case]
fn test_println_simple() {
    println!("test_println_simple output");
}
